/* Operating Systems and Process-Oriented Programming
 * Uppsala University - 2019 Spring
 * Critical sections and mutual exclusion.
 * Multiple threads accessing the a shared bounded buffer.
 * Copyright © 2019 Henrik Alfken(Schulze), Nikos Nikoleris
 Inspired by: https://www.it.uu.se/education/course/homepage/os/vt19/module-4/bounded-buffer/
bounded_buffer.c
 Compile and run by copy-pasting:
make clean && make bin/bounded_buffer && bin/bounded_buffer
 --OR--
make bin/bounded_buffer && bin/bounded_buffer
 */
#include <pthread.h>   // pthread_...
#include <semaphore.h> // sem_...
#include <stdbool.h>   // true, false
#include <stdio.h>     // printf(), fprintf
#include <stdlib.h>    // [s]rand()
#include <unistd.h>    // usleep()
#define BUFFER_SIZE 5
#define PRODUCERS 2
#define CONSUMERS 4
#define ITERATIONS 13
#define PRODUCER_ITERATIONS (ITERATIONS / PRODUCERS)
#define CONSUMER_ITERATIONS (ITERATIONS / CONSUMERS)
#define MAX_SLEEP_TIME 3 // Used for simulating work to be done.
#define SYNCH true     // If true synchronize, if false don't.

typedef struct {
    int value[BUFFER_SIZE];
    int next_in, next_out;
} buffer_t;
buffer_t buffer;
pthread_t consumer_tid[CONSUMERS], producer_tid[PRODUCERS];

/**
 * insert_item - thread safe(?) function to insert items to the bounded buffer
 * @param item the value to be inserted
 * @return 0 in case of sucess -1 otherwise
 */
void insert_item(int item, long int id) {
    // TODO: Check and wait if the buffer is full. Ensure exclusive
     // access to the buffer and use the existing code to remove an item.


    buffer.value[buffer.next_in] = item;
    buffer.next_in = (buffer.next_in + 1) % BUFFER_SIZE;
    printf(" -producer %ld: inserted '%c'\n", id, item);


}

/**
 * remove_item - thread safe(?) function to remove items to the bounded buffer
 * @param item the address of the variable that the removed value will be written
 * @return 0 in case of sucess -1 otherwise
 */
void remove_item(int *item, long int id) {
    // TODO: Check and wait if the buffer is empty. Ensure exclusive
     // access to the buffer and use the existing code to remove an item.


    *item = buffer.value[buffer.next_out];
    buffer.value[buffer.next_out] = -1;
    buffer.next_out = (buffer.next_out + 1) % BUFFER_SIZE;
    printf(" consumer %2ld:  removed '%c' \n", id, *item);


}

/**
 * producer - will iterate PRODUCER_ITERATION times and call the corresponding
 * function to insert an integer to the bounded buffer
 * @param param an integer id of the producer used to distinguish between the
 * multiple producer threads
 * @return nothing
 */
void * producer(void *param) {
  int item, i;
  long int id = (long int)param;
  printf("_vv_ producer $ %ld starts now. _vv_\n",id);
  i = PRODUCER_ITERATIONS;
  while (i--) {
    usleep(1 + rand() % MAX_SLEEP_TIME); // Simulating time needed for work done by this thread.
    item = rand() % 123;
    insert_item(item, id);
  }
  pthread_exit(0);
}

/**
 * consumer - will iterate CONSUMER_ITERATION times and call the corresponding
 * function to remove an integer from the bounded buffer
 * @param param an integer id of the producer used to distinguish between the
 * multiple consumer threads
 * @return nothing
 */
void * consumer(void *param) {
  int item, i;
  long int id = (long int)param;
  printf("_vv_ consumer § %ld starts now. _vv_\n",id);
  i = CONSUMER_ITERATIONS;
  while (i--) {
    usleep(1 + rand() % MAX_SLEEP_TIME); // Simulating time needed for work done by this thread.
    remove_item(&item, id);
  }
  pthread_exit(0);
}

int main() {
  puts("");
  long int i;
  srand(time(NULL));
  /* Create the consumer threads */
  for (i = 0; i < CONSUMERS; i++)
  if (pthread_create(&consumer_tid[i], NULL, consumer, (void *)i) != 0) {
      perror("pthread_create");
      abort();
  }
  /* Create the producer threads */
  for (i = 0; i < PRODUCERS; i++)
  if (pthread_create(&producer_tid[i], NULL, producer, (void *)i) != 0) {
      perror("pthread_create");
      abort();
  }
  /* Wait for them to complete */
  for (i = 0; i < CONSUMERS; i++) {
    if (pthread_join(consumer_tid[i], NULL) != 0) {
      perror("pthread_join");
      abort();
    }
    printf("^^ consumer_%ld_joined successfully. ^^\n", i);
  }
  for (i = 0; i < PRODUCERS; i++) {
    if (pthread_join(producer_tid[i], NULL) != 0) {
      perror("pthread_join");
      abort();
    }
    printf("^^ producer_%ld_joined successfully. ^^\n", i);
  }
  puts("");
  return EXIT_SUCCESS;
}
